@@ -1,14 +1,30 @@
-#Copyright ReportLab Europe Ltd. 2000-2017
+#Copyright ReportLab Europe Ltd. 2000-2019
 #see license.txt for license details
 # $URI:$
-__version__='3.3.0'
+__version__='3.5.34'
 __doc__='''Gazillions of miscellaneous internal utility functions'''
 
-import os, sys, time, types, datetime
-from base64 import decodestring as base64_decodestring, encodestring as base64_encodestring
+import os, sys, time, types, datetime, ast
+from functools import reduce as functools_reduce
+literal_eval = ast.literal_eval
+try:
+    from base64 import decodebytes as base64_decodebytes, encodebytes as base64_encodebytes
+except ImportError:
+    from base64 import decodestring as base64_decodebytes, encodestring as base64_encodebytes
 from reportlab import isPy3
 from reportlab.lib.logger import warnOnce
 from reportlab.lib.rltempfile import get_rl_tempfile, get_rl_tempdir, _rl_getuid
+from . rl_safe_eval import rl_safe_exec, rl_safe_eval, safer_globals
+
+class __UNSET__(object):
+    @staticmethod
+    def __bool__():
+        return False
+    @staticmethod
+    def __str__():
+        return '__UNSET__'
+    __repr__ = __str__
+__UNSET__ = __UNSET__()
 
 try:
     import cPickle as pickle
@@ -57,14 +73,26 @@ if isPy3:
         return md5(s if isBytes(s) else s.encode('utf8')).hexdigest()
 
     def asBytes(v,enc='utf8'):
-        return v if isinstance(v,bytes) else v.encode(enc)
+        if isinstance(v,bytes): return v
+        try:
+            return v.encode(enc)
+        except:
+            annotateException('asBytes(%s,enc=%s) error: ' % (ascii(v),ascii(enc)))
 
     def asUnicode(v,enc='utf8'):
-        return v if isinstance(v,str) else v.decode(enc)
+        if isinstance(v,str): return v
+        try:
+            return v.decode(enc)
+        except:
+            annotateException('asUnicode(%s,enc=%s) error: ' % (ascii(v),ascii(enc)))
 
     def asUnicodeEx(v,enc='utf8'):
-        return v if isinstance(v,str) else v.decode(enc) if isinstance(v,bytes) else str(v)
-
+        if isinstance(v,str): return v
+        try:
+            return v.decode(enc) if isinstance(v,bytes) else str(v)
+        except:
+            annotateException('asUnicodeEx(%s,enc=%s) error: ' % (ascii(v),ascii(enc)))
+        
     def asNative(v,enc='utf8'):
         return asUnicode(v,enc=enc)
 
@@ -121,10 +149,10 @@ if isPy3:
             return str(x).encode(enc)
 
     def encode_label(args):
-        return base64_encodestring(pickle.dumps(args)).strip().decode('latin1')
+        return base64_encodebytes(pickle.dumps(args)).strip().decode('latin1')
 
     def decode_label(label):
-        return pickle.loads(base64_decodestring(label.encode('latin1')))
+        return pickle.loads(base64_decodebytes(label.encode('latin1')))
 
     def rawUnicode(s):
         '''converts first 256 unicodes 1-1'''
@@ -159,7 +187,11 @@ else:
             return join(["%02x" % ord(x) for x in md5(s).digest()], '')
 
     def asBytes(v,enc='utf8'):
-        return v if isinstance(v,str) else v.encode(enc)
+        if isinstance(v,str): return v
+        try:
+            return v.encode(enc)
+        except:
+            annotateException('asBytes(%s,enc=%s) error: ' % (repr(v),repr(enc)))
 
     def asNative(v,enc='utf8'):
         return asBytes(v,enc=enc)
@@ -177,10 +209,18 @@ else:
         return isinstance(v, unicode)
 
     def asUnicode(v,enc='utf8'):
-        return v if isinstance(v,unicode) else v.decode(enc)
+        if isinstance(v,unicode): return v
+        try:
+            return v.decode(enc)
+        except:
+            annotateException('asUnicode(%s,enc=%s) error: ' % (repr(v),repr(enc)))
 
     def asUnicodeEx(v,enc='utf8'):
-        return v if isinstance(v,unicode) else v.decode(enc) if isinstance(v,str) else  unicode(v)
+        if isinstance(v,unicode): return v
+        try:
+            return v.decode(enc) if isinstance(v,str) else str(v)
+        except:
+            annotateException('asUnicodeEx(%s,enc=%s) error: ' % (repr(v),repr(enc)))
 
     def isClass(v):
         return isinstance(v,(types.ClassType,type))
@@ -218,10 +258,10 @@ else:
     from string import letters as ascii_letters, uppercase as ascii_uppercase, lowercase as ascii_lowercase
 
     def encode_label(args):
-        return base64_encodestring(pickle.dumps(args)).strip()
+        return base64_encodebytes(pickle.dumps(args)).strip()
 
     def decode_label(label):
-        return pickle.loads(base64_decodestring(label))
+        return pickle.loads(base64_decodebytes(label))
 
     def rawUnicode(s):
         '''converts first 256 unicodes 1-1'''
@@ -232,6 +272,7 @@ else:
         return s.encode('latin1') if isinstance(s,unicode) else s
 
     def rl_exec(obj, G=None, L=None):
+        '''this is unsafe'''
         if G is None:
             frame = sys._getframe(1)
             G = frame.f_globals
@@ -240,7 +281,7 @@ else:
             del frame
         elif L is None:
             L = G
-        exec("""exec obj in G, L""")
+        exec(obj,G, L)
     rl_exec("""def rl_reraise(t, v, b=None):\n\traise t, v, b\n""")
 
     char2int = ord
@@ -490,22 +531,6 @@ def recursiveImport(modulename, baseDir=None, noCWD=0, debug=0):
     finally:
         sys.path = opath
 
-def recursiveGetAttr(obj, name):
-    "Can call down into e.g. object1.object2[4].attr"
-    return eval(name, obj.__dict__)
-
-def recursiveSetAttr(obj, name, value):
-    "Can call down into e.g. object1.object2[4].attr = value"
-    #get the thing above last.
-    tokens = name.split('.')
-    if len(tokens) == 1:
-        setattr(obj, name, value)
-    else:
-        most = '.'.join(tokens[:-1])
-        last = tokens[-1]
-        parent = recursiveGetAttr(obj, most)
-        setattr(parent, last, value)
-
 def import_zlib():
     try:
         import zlib
@@ -558,9 +583,9 @@ def getArgvDict(**kw):
             elif isinstance(v,int):
                 v = int(av)
             elif isinstance(v,list):
-                v = list(eval(av))
+                v = list(literal_eval(av),{})
             elif isinstance(v,tuple):
-                v = tuple(eval(av))
+                v = tuple(literal_eval(av),{})
             else:
                 raise TypeError("Can't convert string %r to %s" % (av,type(v)))
         return v
@@ -575,7 +600,7 @@ def getArgvDict(**kw):
         handled = 0
         ke = k+'='
         for a in A:
-            if a.find(ke)==0:
+            if a.startswith(ke):
                 av = a[len(ke):]
                 A.remove(a)
                 R[k] = handleValue(v,av,func)
@@ -618,59 +643,73 @@ def open_for_read_by_name(name,mode='b'):
         if 'b' not in mode and os.linesep!='\n': s = s.replace(os.linesep,'\n')
         return getBytesIO(s)
 
-if not isPy3:
-    import urllib2, urllib
-    urlopen=urllib2.urlopen
-    def datareader(url,opener=urllib.URLopener().open):
-        return opener(url).read()
-    del urllib, urllib2
-else:
-    from urllib.request import urlopen
-    from urllib.parse import unquote
-    import base64
-    #copied here from urllib.URLopener.open_data because
-    # 1) they want to remove it
-    # 2) the existing one is borken
-    def datareader(url, unquote=unquote, decodebytes=base64.decodebytes):
-        """Use "data" URL."""
-        # ignore POSTed data
-        #
-        # syntax of data URLs:
-        # dataurl   := "data:" [ mediatype ] [ ";base64" ] "," data
-        # mediatype := [ type "/" subtype ] *( ";" parameter )
-        # data      := *urlchar
-        # parameter := attribute "=" value
+def open_for_read(name,mode='b'):
+    #auto initialized function`
+    if not isPy3:
+        import urllib2, urllib
+        from urlparse import urlparse
+        urlopen=urllib2.urlopen
+        def datareader(url,opener=urllib.URLopener().open):
+            return opener(url).read()
+        del urllib, urllib2
+    else:
+        from urllib.request import urlopen
+        from urllib.parse import unquote, urlparse
+        #copied here from urllib.URLopener.open_data because
+        # 1) they want to remove it
+        # 2) the existing one is borken
+        def datareader(url, unquote=unquote):
+            """Use "data" URL."""
+            # ignore POSTed data
+            #
+            # syntax of data URLs:
+            # dataurl   := "data:" [ mediatype ] [ ";base64" ] "," data
+            # mediatype := [ type "/" subtype ] *( ";" parameter )
+            # data      := *urlchar
+            # parameter := attribute "=" value
+            try:
+                typ, data = url.split(',', 1)
+            except ValueError:
+                raise IOError('data error', 'bad data URL')
+            if not typ:
+                typ = 'text/plain;charset=US-ASCII'
+            semi = typ.rfind(';')
+            if semi >= 0 and '=' not in typ[semi:]:
+                encoding = typ[semi+1:]
+                typ = typ[:semi]
+            else:
+                encoding = ''
+            if encoding == 'base64':
+                # XXX is this encoding/decoding ok?
+                data = base64_decodebytes(data.encode('ascii'))
+            else:
+                data = unquote(data).encode('latin-1')
+            return data
+    from reportlab.rl_config import trustedHosts, trustedSchemes
+    if trustedHosts:
+        import re, fnmatch
+        def xre(s):
+            s = fnmatch.translate(s)
+            return s[4:-3] if s.startswith('(?s:') else s[:-7]
+        trustedHosts = re.compile(''.join(('^(?:',
+                                '|'.join(map(xre,trustedHosts)),
+                                ')\\Z')))
+    def open_for_read(name,mode='b'):
+        '''attempt to open a file or URL for reading'''
+        if hasattr(name,'read'): return name
         try:
-            typ, data = url.split(',', 1)
-        except ValueError:
-            raise IOError('data error', 'bad data URL')
-        if not typ:
-            typ = 'text/plain;charset=US-ASCII'
-        semi = typ.rfind(';')
-        if semi >= 0 and '=' not in typ[semi:]:
-            encoding = typ[semi+1:]
-            typ = typ[:semi]
-        else:
-            encoding = ''
-        if encoding == 'base64':
-            # XXX is this encoding/decoding ok?
-            data = decodebytes(data.encode('ascii'))
-        else:
-            data = unquote(data).encode('latin-1')
-        return data
-    del unquote, base64
-
-def open_for_read(name,mode='b', urlopen=urlopen, datareader=datareader):
-    '''attempt to open a file or URL for reading'''
-    if hasattr(name,'read'): return name
-    try:
-        return open_for_read_by_name(name,mode)
-    except:
-        try:
-            return getBytesIO(datareader(name) if name[:5].lower()=='data:' else urlopen(name).read())
+            return open_for_read_by_name(name,mode)
         except:
-            raise IOError('Cannot open resource "%s"' % name)
-del urlopen, datareader
+            try:
+                if trustedHosts is not None:
+                    purl = urlparse(name)
+                    if purl[0] and not ((purl[0] in ('data','file') or trustedHosts.match(purl[1])) and (purl[0] in trustedSchemes)):
+                        raise ValueError('Attempted untrusted host access')
+                return getBytesIO(datareader(name) if name[:5].lower()=='data:' else urlopen(name).read())
+            except:
+                raise IOError('Cannot open resource "%s"' % name)
+    globals()['open_for_read'] = open_for_read
+    return open_for_read(name,mode)
 
 def open_and_read(name,mode='b'):
     f = open_for_read(name,mode)
@@ -763,6 +802,7 @@ def _isPILImage(im):
 class ImageReader(object):
     "Wraps up either PIL or Java to get data from bitmaps"
     _cache={}
+    _max_image_size = None
     def __init__(self, fileName,ident=None):
         if isinstance(fileName,ImageReader):
             self.__dict__ = fileName.__dict__   #borgize
@@ -810,6 +850,7 @@ class ImageReader(object):
                     #detect which library we are using and open the image
                     if not self._image:
                         self._image = self._read_image(self.fp)
+                        self.check_pil_image_size(self._image)
                     if getattr(self._image,'format',None)=='JPEG': self.jpeg_fh = self._jpeg_fh
                 else:
                     from reportlab.pdfbase.pdfutils import readJPEGInfo
@@ -817,6 +858,10 @@ class ImageReader(object):
                         self._width,self._height,c=readJPEGInfo(self.fp)
                     except:
                         annotateException('\nImaging Library not available, unable to import bitmaps only jpegs\nfileName=%r identity=%s'%(fileName,self.identity()))
+                    size = self._width*self._height*c
+                    if self._max_image_size is not None and size>self._max+image_size:
+                        raise MemoryError('JPEG %s color %s x %s image would use %s > %s bytes'
+                                            %(c,self._width,self._height,size,self._max_image_size))
                     self.jpeg_fh = self._jpeg_fh
                     self._data = self.fp.read()
                     self._dataA=None
@@ -839,6 +884,23 @@ class ImageReader(object):
         else:
             return Image.open(fp)
 
+    @classmethod
+    def check_pil_image_size(cls, im):
+        max_image_size = cls._max_image_size
+        if max_image_size is None: return
+        w, h = im.size
+        m = im.mode
+        size = max(1,((1 if m=='1' else 8*len(m))*w*h)>>3)
+        if size>max_image_size:
+            raise MemoryError('PIL %s %s x %s image would use %s > %s bytes'
+                                            %(m,w,h,size,max_image_size))
+    @classmethod
+    def set_max_image_size(cls,max_image_size=None):
+        cls._max_image_size = max_image_size
+        if max_image_size is not None:
+            from reportlab.rl_config import register_reset
+            register_reset(cls.set_max_image_size)
+
     def _jpeg_fh(self):
         fp = self.fp
         fp.seek(0)
@@ -1375,35 +1437,10 @@ def findInPaths(fn,paths,isfile=True,fail=False):
     if fail: raise ValueError('cannot locate %r with paths=%r' % (fn,paths))
     return fn
 
-def annotateException(msg,enc='utf8'):
+def annotateException(msg,enc='utf8',postMsg='',sep=' '):
     '''add msg to the args of an existing exception'''
-    if not msg: raise
     t,v,b=sys.exc_info()
-    if not hasattr(v,'args'): raise
-    e = -1
-    A = list(v.args)
-    for i,a in enumerate(A):
-        if isinstance(a,str):
-            e = i
-            break
-    if e>=0:
-        if not isPy3:
-            if isUnicode(a):
-                if not isUnicode(msg):
-                    msg=msg.decode(enc)
-            else:
-                if isUnicode(msg):
-                    msg=msg.encode(enc)
-                else:
-                    msg = str(msg)
-        if isinstance(v,IOError) and getattr(v,'strerror',None):
-            v.strerror = msg+'\n'+str(v.strerror)
-        else:
-            A[e] += msg
-    else:
-        A.append(msg)
-    v.args = tuple(A)
-    rl_reraise(t,v,b)
+    rl_reraise(t,t(sep.join((_ for _ in (msg,str(v),postMsg) if _))),b)
 
 def escapeOnce(data):
     """Ensure XML output is escaped just once, irrespective of input
@@ -1516,3 +1553,48 @@ class TimeStamp(object):
             a[2] = a[2].lstrip('0')
             a = ' '.join(a)
         return a
+
+def recursiveGetAttr(obj, name, g=None):
+    "Can call down into e.g. object1.object2[4].attr"
+    if not isStr(name): raise TypeError('invalid reursive acess using %r' % name)
+    name = asNative(name)
+    name = name.strip()
+    if not name: raise ValueError('empty recursive access')
+    dot = '.' if name and name[0] not in '[.(' else ''
+    return rl_safe_eval('obj%s%s'%(dot,name), g={}, l=dict(obj=obj))
+
+def recursiveSetAttr(obj, name, value):
+    "Can call down into e.g. object1.object2[4].attr = value"
+    #get the thing above last.
+    tokens = name.split('.')
+    if len(tokens) == 1:
+        setattr(obj, name, value)
+    else:
+        most = '.'.join(tokens[:-1])
+        last = tokens[-1]
+        parent = recursiveGetAttr(obj, most)
+        setattr(parent, last, value)
+
+def recursiveDelAttr(obj, name):
+    tokens = name.split('.')
+    if len(tokens) == 1:
+        delattr(obj, name)
+    else:
+        most = '.'.join(tokens[:-1])
+        last = tokens[-1]
+        parent = recursiveGetAttr(obj, most)
+        delattr(parent, last)
+
+def yieldNoneSplits(L):
+    '''yield sublists of L separated by None; the Nones disappear'''
+    i = 0
+    n = len(L)
+    while i<n:
+        try:
+            j = L.index(None,i)
+            yield L[i:j]
+            i = j+1
+            if not L: break
+        except ValueError:
+            yield L[i:]
+            break
